#!/usr/bin/env python3
# thoth-glyph
# Copyright(C) 2020 Red Hat, Inc.
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Generate CHANGELOG entries out of commit messages using AI/ML techniques."""

from .lib import classify_message
from .lib import classify_messages
from .lib import classify_by_date
from .lib import classify_by_tag
from .lib import generate_log
from .constants import MLModel
from .constants import Format
from .exceptions import RepositoryNotFoundException
from .exceptions import ModelNotFoundException
from .exceptions import NoMessageEnteredException
from .exceptions import ThothGlyphException

__author__ = "Tushar Sharma <tussharm@redhat.com>"
__title__ = "glyph"
__version__ = "0.1.0"



#!/usr/bin/env python3
# thoth-glyph
# Copyright(C) 2020 Red Hat, Inc.
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Helper functions for the library."""

import logging
import os
from os import path

from pygit2 import Repository
from pygit2 import GIT_SORT_TOPOLOGICAL

import time
import datetime
import sys

from thoth.glyph import __name__
from .exceptions import RepositoryNotFoundException
from .exceptions import NoMessageEnteredException
from .constants import MLModel
from .constants import Format
from .models import FasttextModel
from .formatter import ClusterSimilar

_LOGGER = logging.getLogger(__name__)
DEFAULT_MODEL_PATH = path.join(path.dirname(__file__), "data/model_commits_v2_quant.bin")
CHECK_PHRASES = {"Automatic Updates": ["Automatic Update of dependency"]}


def classify_by_date(path: str, start: str, end: str, model: str):
    start_time = 0
    end_time = sys.maxsize

    if start is not None:
        start_time = int(time.mktime(datetime.datetime.strptime(start, "%Y-%m-%d").timetuple()))

    if end is not None:
        end_time = int(time.mktime(datetime.datetime.strptime(end, "%Y-%m-%d").timetuple()))

    repo_path = os.path.join(path, ".git")
    if os.path.exists(repo_path):
        repo = Repository(repo_path)
    else:
        raise RepositoryNotFoundException

    orig_messages = []
    for commit in repo.walk(repo.head.target, GIT_SORT_TOPOLOGICAL):
        if commit.commit_time > start_time and commit.commit_time < end_time:
            orig_messages.append(commit.message.lower())

    return classify_messages(orig_messages, model)


def classify_by_tag(path: str, start_tag: str, end_tag: str, model: str):
    repo_path = os.path.join(path, ".git")
    if os.path.exists(repo_path):
        repo = Repository(repo_path)
    else:
        raise RepositoryNotFoundException

    start_tag = repo.revparse_single("refs/tags/" + start_tag)

    if end_tag is None:
        end_tag = repo.revparse_single("refs/heads/master")
    else:
        end_tag = repo.revparse_single("refs/tags/" + end_tag)

    orig_messages = []
    walker = repo.walk(end_tag.id, GIT_SORT_TOPOLOGICAL)
    walker.hide(start_tag.id)

    for commit in walker:
        orig_messages.append(commit.message.lower())

    return classify_messages(orig_messages, model)


def classify_messages(messages: list, model: str):
    if messages is None or len(messages) == 0:
        _LOGGER.error("No commits found!")
        return

    if model is None:
        _LOGGER.info("Using default model")
        model = MLModel.DEFAULT

    if model == MLModel.FASTTEXT:
        return FasttextModel.classify_messages(messages)


def classify_message(message: str, model: str) -> str:
    if message is None or message.strip() == "":
        raise NoMessageEnteredException

    if model is None:
        _LOGGER.info("Using default model")
        model = MLModel.DEFAULT

    if model == MLModel.FASTTEXT:
        return FasttextModel.classify_message(message)


def generate_log(messages: list, format: str, model: str):
    check_phrase_dict = {}
    filter_indices = []

    for phrase_heading in CHECK_PHRASES:
        for phrase in CHECK_PHRASES[phrase_heading]:
            for i in range(len(messages)):
                if phrase.lower() in messages[i].lower():
                    filter_indices.append(i)
                    if phrase_heading in check_phrase_dict:
                        temp = check_phrase_dict[phrase_heading]
                        temp.append(messages[i])
                        check_phrase_dict[phrase_heading] = temp
                    else:
                        check_phrase_dict[phrase_heading] = [messages[i]]

    messages = [v for i, v in enumerate(messages) if i not in filter_indices]
    df = classify_messages(messages, model)

    keys = ["features", "corrective", "perfective", "nonfunctional", "unknown"]
    message_dict = {}

    for key in keys:
        message_dict[key] = []

    predicted_labels = list(df["labels_predicted"].astype(str))

    for i in range(len(messages)):
        label = predicted_labels[i]
        temp = message_dict[label]
        temp.append(messages[i])
        message_dict[label] = temp

    # TODO: This tranlational logic is only needed for this specific Fasttext model
    message_dict["Features"] = message_dict.pop("features")
    message_dict["Bug Fixes"] = message_dict.pop("corrective")
    message_dict["Improvements"] = message_dict.pop("perfective")
    message_dict["Non-functional"] = message_dict.pop("nonfunctional")
    message_dict["Other"] = message_dict.pop("unknown")
    for key in check_phrase_dict:
        message_dict[key] = check_phrase_dict[key]

    if model is None:
        _LOGGER.info("Using default format")
        model = Format.DEFAULT

    if format == Format.CLUSTER_SIMILAR:
        return ClusterSimilar.generate_log(message_dict)



#!/usr/bin/env python3
# thoth-glyph
# Copyright(C) 2020 Red Hat, Inc.
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Module containing all supported Machine Learning models."""

from fasttext import load_model
import pandas as pd
from os import path
import logging

_LOGGER = logging.getLogger(__name__)
DEFAULT_FASTTEXT_MODEL_PATH = path.join(path.dirname(__file__), "data/model_commits_v2_quant.bin")


class FasttextModel:

    @staticmethod
    def classify_message(message: str) -> str:
        _LOGGER.info("Model Path : " + DEFAULT_FASTTEXT_MODEL_PATH)
        classifier = load_model(DEFAULT_FASTTEXT_MODEL_PATH)
        label = classifier.predict(message.lower())
        label_string = str(label[0][0])[9:]
        return label_string

    @staticmethod
    def classify_messages(messages: list):
        df = pd.DataFrame(messages, columns=["message"])
        df = df.replace("\n", "", regex=True)
        commits = list(df["message"].astype(str))
        _LOGGER.info("Model Path : " + DEFAULT_FASTTEXT_MODEL_PATH)
        classifier = load_model(DEFAULT_FASTTEXT_MODEL_PATH)
        labels = classifier.predict(commits)
        res = list(zip(*labels))
        res_list = [x[0] for x in res]
        lst2 = [item[0] for item in res_list]
        df["labels_predicted"] = lst2
        df["labels_predicted"] = df["labels_predicted"].map(lambda x: x[9:])
        _LOGGER.info(str(len(messages)) + " commits classified")
        return df



#!/usr/bin/env python3
# thoth-glyph
# Copyright(C) 2020 Red Hat, Inc.
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Enums used in Glyph."""

from enum import Enum


class _ExtendedEnum(Enum):
    """A custom enum with extended functionality."""

    @classmethod
    def by_name(cls, name: str) -> Enum:
        """Retrieve enum based on its name."""
        try:
            return cls.__members__[name.upper()]
        except KeyError as exc:
            raise ValueError("Unknown value for type {}, available: {}", cls, list(cls.__members__.values()),) from exc


class MLModel(_ExtendedEnum):
    """Supported ML Classifiers."""

    FASTTEXT = 0
    DEFAULT = FASTTEXT


class Format(_ExtendedEnum):
    """Supported changelog formats."""

    CLUSTER_SIMILAR = 0
    DEFAULT = CLUSTER_SIMILAR



#!/usr/bin/env python3
# thoth-glyph
# Copyright(C) 2020 Tushar Sharma <tussharm@redhat.com>
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Glyph's CLI Interface."""

import logging
from typing import Optional

import click
from thoth.common import init_logging

from thoth.glyph import __title__
from thoth.glyph import __version__ as glyph_version
from thoth.glyph import classify_message
from thoth.glyph import classify_by_date
from thoth.glyph import classify_by_tag
from thoth.glyph import MLModel

init_logging()
_LOGGER = logging.getLogger(__title__)


def _print_version(ctx: click.Context, _, value: str):
    """Print glyph version and exit."""
    if not value or ctx.resilient_parsing:
        return

    click.echo(glyph_version)
    ctx.exit()


@click.group()
@click.pass_context
@click.option(
    "-v", "--verbose", is_flag=True, envvar="THOTH_ADVISER_DEBUG", help="Be verbose about what's going on.",
)
@click.option(
    "--version",
    is_flag=True,
    is_eager=True,
    callback=_print_version,
    expose_value=False,
    help="Print adviser version and exit.",
)
def cli(ctx: Optional[click.Context] = None, verbose: bool = False):
    """Glyph command line interface."""
    if ctx:
        ctx.auto_envvar_prefix = "GLYPH"

    if verbose:
        _LOGGER.setLevel(logging.DEBUG)

    _LOGGER.debug("Debug mode is on")
    _LOGGER.info("Version: %s", glyph_version)


@cli.command()
@click.option(
    "--message", "-m", type=str, required=True, help="Commit message to be classified",
)
@click.option(
    "--model",
    default=MLModel.DEFAULT.name.lower(),
    type=click.Choice([e.name.lower() for e in MLModel]),
    help="Type of classifer",
)
def classify(message: str, model: str) -> None:
    """Generate CHANGELOG entries from the current Git project."""
    _LOGGER.info("Classifying commit")
    model = MLModel.by_name(model)
    print("Label : " + classify_message(message, model))


@cli.command("classify-repo")
@click.option("--path", "-p", type=str, required=True, help="Path to Git repository")
@click.option("--start", type=str, help="Starting date")
@click.option("--end", type=str, help="End date")
@click.option("--output", type=str, help="Generated output file")
@click.option(
    "--model",
    default=MLModel.DEFAULT.name.lower(),
    type=click.Choice([e.name.lower() for e in MLModel]),
    help="Type of classifer",
)
def classifybydate(path: str, start: str, end: str, output: str, model: str) -> None:
    _LOGGER.info("Classifying commits in the given date-range")
    model = MLModel.by_name(model)
    df = classify_by_date(path, start, end, model)
    if output is None:
        print(df)
    else:
        df.to_csv(output, sep="\t")


@cli.command("classify-repo-by-tag")
@click.option("--path", "-p", type=str, required=True, help="Path to Git repository")
@click.option("--start_tag", type=str, required=True, help="Start tag")
@click.option("--end_tag", type=str, help="End tag")
@click.option("--output", type=str, help="Generated output file")
@click.option(
    "--model",
    default=MLModel.DEFAULT.name.lower(),
    type=click.Choice([e.name.lower() for e in MLModel]),
    help="Type of classifer",
)
def classifybytag(path: str, start_tag: str, end_tag: str, output: str, model: str) -> None:
    _LOGGER.info("Classifying commits between given tags")
    model = MLModel.by_name(model)
    df = classify_by_tag(path, start_tag, end_tag, model)
    if output is None:
        print(df)
    else:
        df.to_csv(output, sep="\t")


__name__ == "__main__" and cli()



#!/usr/bin/env python3
# thoth-glyph
# Copyright(C) 2020 Red Hat, Inc.
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Module containing all supported formatting options for the changelog file."""


class ClusterSimilar:

    @staticmethod
    def generate_log(message_dict: dict):
        changelog = []
        for key in message_dict:
            if message_dict[key] != None and len(message_dict[key]) > 0:
                changelog.append("### " + key)
                for message in message_dict[key]:
                    changelog.append("* " + message)
        return changelog
